from __future__ import annotations

import bz2
import gzip
import logging
import lzma
import os
import shutil
import tarfile
import tempfile
from contextlib import suppress
from os import PathLike
from pathlib import Path
from typing import IO, Any, Literal, Union, cast
from urllib.parse import urlparse
from zipfile import ZipFile, is_zipfile

logger = logging.getLogger(__name__)


OpenTextModeUpdating = Literal[
    "r+",
    "+r",
    "rt+",
    "r+t",
    "+rt",
    "tr+",
    "t+r",
    "+tr",
    "w+",
    "+w",
    "wt+",
    "w+t",
    "+wt",
    "tw+",
    "t+w",
    "+tw",
    "a+",
    "+a",
    "at+",
    "a+t",
    "+at",
    "ta+",
    "t+a",
    "+ta",
    "x+",
    "+x",
    "xt+",
    "x+t",
    "+xt",
    "tx+",
    "t+x",
    "+tx",
]
OpenTextModeWriting = Literal[
    "w",
    "wt",
    "tw",
    "a",
    "at",
    "ta",
    "x",
    "xt",
    "tx",
]
OpenTextModeReading = Literal[
    "r",
    "rt",
    "tr",
    "U",
    "rU",
    "Ur",
    "rtU",
    "rUt",
    "Urt",
    "trU",
    "tUr",
    "Utr",
]
OpenTextMode = Union[OpenTextModeUpdating, OpenTextModeWriting, OpenTextModeReading]
OpenBinaryModeUpdating = Literal[
    "rb+",
    "r+b",
    "+rb",
    "br+",
    "b+r",
    "+br",
    "wb+",
    "w+b",
    "+wb",
    "bw+",
    "b+w",
    "+bw",
    "ab+",
    "a+b",
    "+ab",
    "ba+",
    "b+a",
    "+ba",
    "xb+",
    "x+b",
    "+xb",
    "bx+",
    "b+x",
    "+bx",
]
OpenBinaryModeWriting = Literal[
    "wb",
    "bw",
    "ab",
    "ba",
    "xb",
    "bx",
]
OpenBinaryModeReading = Literal[
    "rb",
    "br",
    "rbU",
    "rUb",
    "Urb",
    "brU",
    "bUr",
    "Ubr",
]
OpenBinaryMode = Union[OpenBinaryModeUpdating, OpenBinaryModeReading, OpenBinaryModeWriting]


def remove_file_or_directory(path: str | PathLike) -> None:
    with suppress(FileNotFoundError):
        path = Path(path)
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()


def is_archive_file(filename: str | PathLike) -> bool:
    if not Path(filename).is_file():
        return False
    return is_zipfile(filename) or tarfile.is_tarfile(filename)


def extract_archive_file(
    source_path: str | PathLike,
    target_path: str | PathLike,
) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        if is_zipfile(source_path):
            with ZipFile(source_path, "r") as zip_file:
                zip_file.extractall(temp_dir)
        elif tarfile.is_tarfile(source_path):
            with tarfile.open(source_path) as tar_file:
                tar_file.extractall(temp_dir)
        os.replace(temp_dir, target_path)


def extract_path(filename: str | PathLike) -> str:
    parsed = urlparse(str(filename))
    return parsed.path


def is_local(url_or_filename: str | PathLike) -> bool:
    if isinstance(url_or_filename, Path):
        return True

    parsed = urlparse(str(url_or_filename))
    if parsed.scheme in ("", "file", "osfs"):
        return True

    return False


def get_parent_path_and_filename(path: str | PathLike) -> tuple[str, str]:
    if isinstance(path, Path):
        parent = str(path.parent)
        name = str(path.name)
        return parent, name

    path = str(path)
    parsed_url = urlparse(path)
    scheme = parsed_url.scheme
    netloc = parsed_url.netloc
    path = parsed_url.path

    splitted = path.rsplit("/", 1)
    if len(splitted) == 2:
        parent, name = splitted
    else:
        parent = ""
        name = path

    if scheme and netloc:
        parent = f"{scheme}://{netloc}/{parent}"

    return parent, name


def sizeof_fmt(num: int | float, suffix: str = "", dividor: int | float = 1024) -> str:
    for unit in ("", "K", "M", "G", "T", "P", "E", "Z"):
        if abs(num) < dividor:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= dividor
    return "%.1f%s%s" % (num, "Yi", suffix)


DecompressOption = Literal["none", "force", "auto"]


def xopen(
    file: str | PathLike,
    mode: str = "r",
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
    *,
    decompress: DecompressOption = "auto",
) -> IO[Any]:
    if decompress == "none":
        return open(file, mode, buffering, encoding, errors, newline=newline)

    if Path(file).exists():
        with suppress(gzip.BadGzipFile):
            with gzip.open(file, "r", encoding=encoding, errors=errors, newline=newline) as gzipfile:
                gzipfile.read(1)
                gzipfile.seek(0)
            return cast(IO[Any], gzip.open(file, mode, encoding=encoding, errors=errors, newline=newline))

        with suppress(lzma.LZMAError):
            with lzma.open(file, "r", encoding=encoding, errors=errors, newline=newline) as lzmafile:
                lzmafile.read(1)
                lzmafile.seek(0)
            return cast(IO[Any], lzma.open(file, mode, encoding=encoding, errors=errors, newline=newline))

        with suppress(IOError):
            with bz2.open(file, "r", encoding=encoding, errors=errors, newline=newline) as bz2file:
                bz2file.read(1)
                bz2file.seek(0)
            return cast(IO[Any], bz2.open(file, mode, encoding=encoding, errors=errors, newline=newline))

        raise ValueError(f"Failed to open with decompress: {file}")

    if str(file).endswith(".gz"):
        return cast(IO[Any], gzip.open(file, mode, encoding=encoding, errors=errors, newline=newline))

    if str(file).endswith((".xz", ".lzma")):
        return cast(IO[Any], lzma.open(file, mode, encoding=encoding, errors=errors, newline=newline))

    if str(file).endswith(".bz2"):
        return cast(IO[Any], bz2.open(file, mode, encoding=encoding, errors=errors, newline=newline))

    if decompress == "auto":
        return open(file, mode, buffering, encoding, errors, newline=newline)

    raise ValueError(f"Unknown compression type for file: {file}")



from __future__ import annotations

from importlib.metadata import version
from os import PathLike
from pathlib import Path
from typing import IO, Any, BinaryIO, ContextManager, TextIO, overload

from minato.cache import Cache
from minato.config import Config
from minato.filesystems import FileSystem
from minato.minato import Minato
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode

__version__ = version("minato")
__all__ = [
    "Cache",
    "Config",
    "FileSystem",
    "Minato",
    "cached_path",
    "download",
    "exists",
    "open",
    "upload",
]


@overload
def open(
    url_or_filename: str | PathLike,
    mode: OpenTextMode = ...,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
    *,
    extract: bool = ...,
    decompress: DecompressOption = ...,
    auto_update: bool | None = ...,
    use_cache: bool = ...,
    force_download: bool = ...,
    force_extract: bool = ...,
    cache_root: str | PathLike | None = ...,
    expire_days: int | None = ...,
    retry: bool = ...,
) -> ContextManager[TextIO]:
    ...


@overload
def open(
    url_or_filename: str | PathLike,
    mode: OpenBinaryMode,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
    *,
    extract: bool = ...,
    decompress: DecompressOption = ...,
    auto_update: bool | None = ...,
    use_cache: bool = ...,
    force_download: bool = ...,
    force_extract: bool = ...,
    cache_root: str | PathLike | None = ...,
    expire_days: int | None = ...,
    retry: bool = ...,
) -> ContextManager[BinaryIO]:
    ...


@overload
def open(
    url_or_filename: str | PathLike,
    mode: str,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
    *,
    extract: bool = ...,
    decompress: DecompressOption = ...,
    auto_update: bool | None = ...,
    use_cache: bool = ...,
    force_download: bool = ...,
    force_extract: bool = ...,
    cache_root: str | PathLike | None = ...,
    expire_days: int | None = ...,
    retry: bool = ...,
) -> ContextManager[IO[Any]]:
    ...


def open(
    url_or_filename: str | PathLike,
    mode: str = "r",
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
    *,
    extract: bool = False,
    decompress: DecompressOption = "none",
    auto_update: bool | None = None,
    use_cache: bool = True,
    force_download: bool = False,
    force_extract: bool = False,
    cache_root: str | PathLike | None = None,
    expire_days: int | None = None,
    retry: bool = True,
) -> ContextManager[IO[Any]]:
    config = Config.load(
        cache_root=cache_root,
    )

    return Minato(config).open(
        url_or_filename,
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        extract=extract,
        decompress=decompress,
        auto_update=auto_update,
        use_cache=use_cache,
        force_download=force_download,
        force_extract=force_extract,
        expire_days=expire_days,
        retry=retry,
    )


def cached_path(
    url_or_filename: str | PathLike,
    extract: bool = False,
    auto_update: bool | None = None,
    force_download: bool = False,
    force_extract: bool = False,
    retry: bool = True,
    cache_root: str | PathLike | None = None,
    expire_days: int | None = None,
) -> Path:
    config = Config.load(
        cache_root=cache_root,
    )

    return Minato(config).cached_path(
        url_or_filename,
        extract=extract,
        auto_update=auto_update,
        force_download=force_download,
        force_extract=force_extract,
        expire_days=expire_days,
        retry=retry,
    )


def download(url: str, filename: str | PathLike) -> None:
    filename = Path(filename)
    Minato.download(url, filename)


def upload(filename: str | PathLike, url: str) -> None:
    filename = Path(filename)
    Minato.upload(filename, url)


def delete(url_or_filename: str | PathLike) -> None:
    Minato.delete(url_or_filename)


def exists(url_or_filename: str | PathLike) -> bool:
    return Minato.exists(url_or_filename)



from __future__ import annotations

import logging
import os
import re
import tempfile
import threading
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import IO, Any, BinaryIO, ContextManager, Iterator, TextIO, overload

from minato.common import Progress
from minato.filesystems.filesystem import FileSystem
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode, sizeof_fmt, xopen

try:
    import boto3
except ModuleNotFoundError:
    boto3 = None  # type: ignore[assignment]


logger = logging.getLogger(__name__)


@FileSystem.register(["s3"])
class S3FileSystem(FileSystem):
    def __init__(self, url_or_filename: str | PathLike) -> None:
        if boto3 is None:
            raise ModuleNotFoundError(
                "S3FileSystem is not available. Please make sure that " "boto3 is successfully installed."
            )

        super().__init__(url_or_filename)
        self._tlocal = threading.local()

        self._bucket_name = self._url.netloc or ""
        self._key = re.sub(r"^/", "", self._url.path)

        self._aws_access_key_id = self._url.username or os.environ.get("AWS_ACCESS_KEY_ID")
        self._aws_secret_access_key = self._url.password or os.environ.get("AWS_SECRET_ACCESS_KEY")
        self._endpoint_url = self._url.get_query("endpoint_url")
        self._region_name = self._url.get_query("region")

    def _get_resource(self):  # type: ignore
        if not hasattr(self._tlocal, "resource"):
            self._tlocal.resource = boto3.resource(
                "s3",
                region_name=self._region_name,
                aws_access_key_id=self._aws_access_key_id,
                aws_secret_access_key=self._aws_secret_access_key,
                endpoint_url=self._endpoint_url,
            )
        return self._tlocal.resource

    def _get_client(self):  # type: ignore
        if not hasattr(self._tlocal, "client"):
            self._tlocal.client = boto3.client(
                "s3",
                region_name=self._region_name,
                aws_access_key_id=self._aws_access_key_id,
                aws_secret_access_key=self._aws_secret_access_key,
                endpoint_url=self._endpoint_url,
            )
        return self._tlocal.client

    def _download_fileobj(self, key: str, fp: IO[Any]) -> None:
        client = self._get_client()  # type: ignore
        client.download_fileobj(self._bucket_name, key, fp)

    def _upload_fileobj(self, fp: IO[Any], key: str) -> None:
        client = self._get_client()  # type: ignore
        client.upload_fileobj(fp, self._bucket_name, key)

    def exists(self) -> bool:
        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)
        objects = list(bucket.objects.filter(Prefix=self._key))
        return len(objects) > 0

    def download(self, path: str | PathLike) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        path = Path(path)

        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)
        objects = [obj for obj in bucket.objects.filter(Prefix=self._key) if not obj.key.endswith("/")]
        total = sum(obj.size for obj in objects)

        logger.debug("%s file(s) (%sB) will be downloaded to %s.", len(objects), total, path)

        path = path / os.path.basename(self._key) if path.is_dir() else path

        with Progress[int](
            total,
            unit="iB",
            desc="downloading from s3",
            sizeof_formatter=sizeof_fmt,
        ) as progress:
            for obj in objects:
                relprefix = os.path.relpath(obj.key, self._key)
                file_path = path / relprefix
                os.makedirs(file_path.parent, exist_ok=True)
                bucket.download_file(obj.key, str(file_path), Callback=progress.update)

    def upload(self, path: str | PathLike) -> None:
        path = Path(str(path)).absolute()

        prefix = self._key
        if prefix.endswith("/"):
            prefix = os.path.join(prefix, path.name)

        filenames = [subpath for subpath in path.glob("**/*") if subpath.is_file()] if path.is_dir() else [path]

        total = sum(filename.stat().st_size for filename in filenames)

        logger.info("%s file(s) (%sB) will be uploaded to %s", len(filenames), total, self._url)

        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)

        with Progress[int](
            total,
            unit="iB",
            desc="uploading to s3",
            sizeof_formatter=sizeof_fmt,
        ) as progress:
            for filename in filenames:
                if filename != path:
                    relpath = os.path.relpath(filename, path)
                    key = os.path.join(prefix, relpath)
                else:
                    key = prefix
                bucket.upload_file(str(filename), key, Callback=progress.update)

    def delete(self) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)
        bucket.objects.filter(Prefix=self._key).delete()

    def get_version(self) -> str | None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        resource = self._get_resource()  # type: ignore
        bucket = resource.Bucket(self._bucket_name)
        objects = list(bucket.objects.filter(Prefix=self._key))
        etags = [str(obj.e_tag).strip('"') for obj in objects if obj.e_tag]
        return ".".join(sorted(etags)) if etags else None

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[TextIO]:
        ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[BinaryIO]:
        ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[IO[Any]]:
        ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        decompress: DecompressOption = "none",
    ) -> ContextManager[IO[Any]]:
        @contextmanager
        def _open(
            mode: str,
            buffering: int,
            encoding: str | None,
            errors: str | None,
            newline: str | None,
        ) -> Iterator[IO[Any]]:
            if "x" in mode and self.exists():
                raise FileExistsError(self._url.raw)

            suffix = "-" + (self._url.path.split("/")[-1] if self._url.path else "")
            local_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)

            try:
                if "r" in mode or "a" in mode or "+" in mode:
                    if not self.exists():
                        raise FileNotFoundError(self._url.raw)
                    self._download_fileobj(self._key, local_file)

                local_file.close()

                with xopen(
                    local_file.name,
                    mode=mode,
                    buffering=buffering,
                    encoding=encoding,
                    errors=errors,
                    newline=newline,
                    decompress=decompress,
                ) as fp:
                    yield fp

                if "w" in mode or "a" in mode or "+" in mode or "x" in mode:
                    with open(local_file.name, "rb") as fp:
                        self._upload_fileobj(fp, self._key)
            finally:
                local_file.close()
                os.remove(local_file.name)

        return _open(mode, buffering, encoding, errors, newline)



class MinatoException(Exception):
    """MinatoException"""


class ConfigurationError(MinatoException):
    """ConfigurationError"""


class CacheNotFoundError(MinatoException):
    """CacheNotFoundError"""


class InvalidCacheStatus(MinatoException):
    """InvalidCacheStatus"""


class CacheAlreadyExists(MinatoException):
    """CacheAlreadyExists"""



from __future__ import annotations

import logging
from collections.abc import Callable
from os import PathLike
from typing import IO, Any, BinaryIO, ContextManager, TextIO, Type, overload
from urllib.parse import urlparse

from minato.common import URL
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode

logger = logging.getLogger(__name__)


@overload
def open_file(
    url_or_filename: str | PathLike,
    mode: OpenTextMode = ...,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
    *,
    decompress: DecompressOption = ...,
) -> ContextManager[TextIO]:
    ...


@overload
def open_file(
    url_or_filename: str | PathLike,
    mode: OpenBinaryMode,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
    *,
    decompress: DecompressOption = ...,
) -> ContextManager[BinaryIO]:
    ...


@overload
def open_file(
    url_or_filename: str | PathLike,
    mode: str,
    buffering: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
    *,
    decompress: DecompressOption = ...,
) -> ContextManager[IO[Any]]:
    ...


def open_file(
    url_or_filename: str | PathLike,
    mode: str = "r",
    buffering: int = -1,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
    *,
    decompress: DecompressOption = "none",
) -> ContextManager[IO[Any]]:
    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    return filesystem.open_file(
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        decompress=decompress,
    )


def exists(url_or_filename: str | PathLike) -> bool:
    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    return filesystem.exists()


def download(
    url_or_filename: str | PathLike,
    download_path: str | PathLike,
) -> None:
    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    filesystem.download(download_path)


def upload(source: str | PathLike, target: str | PathLike) -> None:
    target = str(target)
    filesystem = FileSystem.by_url(target)
    filesystem.upload(source)


def delete(url_or_filename: str | PathLike) -> None:
    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    filesystem.delete()


def get_version(
    url_or_filename: str | PathLike,
) -> str | None:
    url = str(url_or_filename)
    filesystem = FileSystem.by_url(url)
    return filesystem.get_version()


class FileSystem:
    registry: dict[str, Type["FileSystem"]] = {}

    @classmethod
    def register(cls, schemes: list[str]) -> Callable[[Type[FileSystem]], Type[FileSystem]]:
        def decorator(subclass: Type[FileSystem]) -> Type[FileSystem]:
            for scheme in schemes:
                FileSystem.registry[scheme] = subclass
            return subclass

        return decorator

    @classmethod
    def by_url(cls, url: str) -> FileSystem:
        parsed_url = urlparse(url)
        scheme = parsed_url.scheme
        if scheme not in FileSystem.registry:
            schemes = ", ".join(FileSystem.registry.keys())
            raise KeyError(f"Invalid scheme: {scheme} (not in {schemes})")
        subclass = FileSystem.registry[scheme]
        logger.debug("Infer file system of %s from url: %s", subclass.__name__, url)
        return subclass(url)

    def __init__(self, url_or_filename: str | PathLike) -> None:
        self._url = URL(str(url_or_filename))

    def exists(self) -> bool:
        raise NotImplementedError

    def download(self, path: str | PathLike) -> None:
        raise NotImplementedError

    def upload(self, path: str | PathLike) -> None:
        raise NotImplementedError

    def delete(self) -> None:
        raise NotImplementedError

    def get_version(self) -> str | None:
        raise NotImplementedError

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[TextIO]:
        ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[BinaryIO]:
        ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[IO[Any]]:
        ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        decompress: DecompressOption = "none",
    ) -> ContextManager[IO[Any]]:
        raise NotImplementedError



from __future__ import annotations

from urllib.parse import parse_qs, urlparse


class URL:
    def __init__(self, url: str) -> None:
        self._raw = url

        self._parse_result = urlparse(url)
        self._queries = parse_qs(self._parse_result.query)

    def __repr__(self) -> str:
        return self._raw

    @property
    def raw(self) -> str:
        return self._raw

    @property
    def path(self) -> str:
        return self._parse_result.path

    @property
    def scheme(self) -> str:
        return self._parse_result.scheme

    @property
    def username(self) -> str | None:
        return self._parse_result.username

    @property
    def password(self) -> str | None:
        return self._parse_result.password

    @property
    def hostname(self) -> str | None:
        return self._parse_result.hostname

    @property
    def netloc(self) -> str | None:
        return self._parse_result.netloc

    def get_queries(self, key: str) -> list[str] | None:
        return self._queries.get(key)

    def get_query(self, key: str) -> str | None:
        values = self.get_queries(key)
        if not values:
            return None
        return values[0]



from __future__ import annotations

import logging
import os
import re
import tempfile
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import IO, Any, BinaryIO, ContextManager, Iterator, TextIO, overload

from minato.common import Progress
from minato.filesystems.filesystem import FileSystem
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode, sizeof_fmt, xopen

try:
    import google.cloud.storage as gcs
    from google.cloud.storage import Blob, Client
except ImportError:
    gcs = None
    Blob, Client = None, None


logger = logging.getLogger(__name__)


@FileSystem.register(["gs", "gcs"])
class GCSFileSystem(FileSystem):
    def __init__(self, url_or_filename: str | PathLike) -> None:
        if gcs is None:
            raise ModuleNotFoundError(
                "GCSFileSystem is not available. Please make sure that "
                "google-cloud-storage is successfully installed."
            )

        super().__init__(url_or_filename)
        self._bucket_name = self._url.netloc or ""
        self._key = re.sub(r"^/", "", self._url.path)

        self._project = self._url.get_query("project")
        self._api_endpoint = self._url.get_query("api_endpoint")

        self._client = Client()
        if self._project:
            self._client.project = self._project
        if self._api_endpoint:
            self._client.client_options = {"api_endpoint": self._api_endpoint}

    def _get_blob(self, blob_name: str) -> Blob:
        bucket = self._client.bucket(self._bucket_name)
        blob = bucket.blob(blob_name)
        return blob

    def _get_url_from_blob(self, blob: Blob) -> str:
        return "gs://{self._bucket_name}/{blob.name}"

    def _download_fileobj(self, key: str, fp: IO[Any]) -> None:
        blob = self._get_blob(key)
        blob.download_to_file(fp)

    def _upload_fileobj(self, fp: IO[Any], key: str) -> None:
        blob = self._get_blob(key)
        blob.upload_from_file(fp)

    def exists(self) -> bool:
        client = self._client
        bucket = client.bucket(self._bucket_name)
        blobs = list(bucket.list_blobs(prefix=self._key))
        return len(blobs) > 0

    def download(self, path: str | PathLike) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        if not isinstance(path, Path):
            path = Path(path)

        client = self._client
        bucket = client.bucket(self._bucket_name)
        blobs = [blob for blob in bucket.list_blobs(prefix=self._key) if not blob.name.endswith("/")]
        total = sum(blob.size for blob in blobs if blob.size)

        logger.debug("%s file(s) (%sB) will be downloaded to %s.", len(blobs), total, path)
        with Progress[int](
            total,
            unit="iB",
            desc="downloading from gcs",
            sizeof_formatter=sizeof_fmt,
        ) as progress:
            for blob in blobs:
                relpath = os.path.relpath(blob.name, self._key)
                file_path = path / relpath
                os.makedirs(file_path.parent, exist_ok=True)
                blob.download_to_filename(str(file_path))
                progress.update(blob.size or 0)

    def upload(self, path: str | PathLike) -> None:
        path = Path(path)

        if not path.exists():
            raise FileNotFoundError(path)

        prefix = self._key
        if prefix.endswith("/"):
            prefix = os.path.join(prefix, path.name)

        filenames = [subpath for subpath in path.glob("**/*") if subpath.is_file()] if path.is_dir() else [path]

        total = sum(filename.stat().st_size for filename in filenames)

        logger.debug("%s file(s) (%sB) will be uploaded to %s", len(filenames), total, self._url)

        client = self._client
        bucket = client.bucket(self._bucket_name)

        with Progress[int](
            total,
            unit="iB",
            desc="uploading from gcs",
            sizeof_formatter=sizeof_fmt,
        ) as progress:
            for filename in filenames:
                if filename != path:
                    relpath = os.path.relpath(filename, path)
                    key = os.path.join(prefix, relpath)
                else:
                    key = prefix

                blob = bucket.blob(key)
                blob.upload_from_filename(str(filename))
                progress.update(filename.stat().st_size)

    def delete(self) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        client = self._client
        bucket = client.bucket(self._bucket_name)
        blobs = list(bucket.list_blobs(prefix=self._key))
        for blob in blobs:
            blob.delete()

    def get_version(self) -> str | None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        client = self._client
        bucket = client.bucket(self._bucket_name)
        blobs = list(bucket.list_blobs(prefix=self._key))
        hashes = [str(blob.md5_hash) for blob in blobs if blob.md5_hash and not blob.name.endswith("/")]
        return ".".join(sorted(hashes)) if hashes else None

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[TextIO]:
        ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[BinaryIO]:
        ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[IO[Any]]:
        ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        decompress: DecompressOption = "none",
    ) -> ContextManager[IO[Any]]:
        @contextmanager
        def _open(
            mode: str,
            buffering: int,
            encoding: str | None,
            errors: str | None,
            newline: str | None,
        ) -> Iterator[IO[Any]]:
            if "x" in mode and self.exists():
                raise FileExistsError(self._url.raw)

            suffix = "-" + (self._url.path.split("/")[-1] if self._url.path else "")
            local_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)

            try:
                if "r" in mode or "a" in mode or "+" in mode:
                    if not self.exists():
                        raise FileNotFoundError(self._url.raw)
                    self._download_fileobj(self._key, local_file)

                local_file.close()
                with xopen(
                    local_file.name,
                    mode=mode,
                    buffering=buffering,
                    encoding=encoding,
                    errors=errors,
                    newline=newline,
                    decompress=decompress,
                ) as fp:
                    yield fp

                if "w" in mode or "a" in mode or "+" in mode or "x" in mode:
                    with open(local_file.name, "rb") as fp:
                        self._upload_fileobj(fp, self._key)

            finally:
                local_file.close()
                os.remove(local_file.name)

        return _open(mode, buffering, encoding, errors, newline)



import argparse
import sys
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.common import Selector
from minato.config import Config
from minato.minato import Minato


@Subcommand.register(
    name="cache",
    description="cache remote file and return cached local file path",
)
class CacheCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "url",
            type=str,
            nargs="?",
            default=None,
            help="specify file path or url",
        )
        self.parser.add_argument(
            "--extract",
            action="store_true",
            help="extract archive file",
        )
        self.parser.add_argument(
            "--auto-update",
            action="store_true",
            help="download new version if available",
        )
        self.parser.add_argument(
            "--force-extract",
            action="store_true",
            help="force to extract the specified archive file",
        )
        self.parser.add_argument(
            "--force-download",
            action="store_true",
            help="force to download the specified file whether the cache exists or not",
        )
        self.parser.add_argument(
            "--not-retry",
            action="store_false",
            help="do not retry to download the file even if " "the previous download operation is failed",
        )
        self.parser.add_argument(
            "--expire-days",
            type=int,
            default=None,
            help="specify expire days of the cache",
        )
        self.parser.add_argument(
            "--root",
            type=Path,
            default=None,
            help="specify the root directory path of cached data",
        )

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
        )
        minato = Minato(config)

        if args.url:
            url = args.url
        else:
            candidate_urls = [cached_file.url for cached_file in minato.cache.all()]
            url = Selector(config.selector_command)(candidate_urls)

        if not url:
            print("URL NOT FOUND", file=sys.stderr)
            sys.exit(1)

        cached_path = minato.cached_path(
            url,
            extract=args.extract,
            auto_update=args.auto_update,
            expire_days=args.expire_days,
            force_extract=args.force_extract,
            force_download=args.force_download,
            retry=args.not_retry,
        )
        print(cached_path)



from __future__ import annotations

import functools
import os
from contextlib import contextmanager
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import IO, Any, BinaryIO, ContextManager, Iterator, TextIO, overload

from minato.common import Progress
from minato.filesystems.filesystem import FileSystem
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode, sizeof_fmt

try:
    from huggingface_hub import HfFileSystem
except ModuleNotFoundError:
    HfFileSystem = None


logger = getLogger(__name__)


@FileSystem.register(["hf"])
class HuggingfaceHubFileSystem(FileSystem):
    @functools.cached_property
    def _hf(self) -> "HfFileSystem":
        if HfFileSystem is None:
            raise ModuleNotFoundError(
                "HuggingfaceHubFileSystem is not available. Please make sure that "
                "huggingface-hub is successfully installed."
            )
        return HfFileSystem()

    def exists(self) -> bool:
        return bool(self._hf.exists(self._url.raw))

    def download(self, path: str | PathLike) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        path = Path(path)

        root = f"{self._url.netloc}/{self._url.path}"
        if self._hf.isdir(self._url.raw):
            paths = self._hf.expand_path(f"{self._url.raw}/**")
        else:
            paths = self._hf.expand_path(self._url.raw)

        paths = [path for path in paths if not self._hf.isdir(path)]
        total = sum(self._hf.size(path) for path in paths)

        logger.debug("%s file(s) (%sB) will be downloaded to %s.", len(paths), total, path)

        path = path / os.path.basename(root) if path.is_dir() else path

        with Progress[int](
            total,
            unit="iB",
            desc="downloading from s3",
            sizeof_formatter=sizeof_fmt,
        ) as progress:
            for source_path in paths:
                relprefix = os.path.relpath(source_path, root)
                target_path = path / relprefix
                target_path.parent.mkdir(parents=True, exist_ok=True)
                with self._hf.open(source_path, "rb") as src, target_path.open("wb") as dst:
                    while True:
                        chunk = src.read(1024)
                        if not chunk:
                            break
                        dst.write(chunk)
                        progress.update(len(chunk))

    def upload(self, path: str | PathLike) -> None:
        path = Path(str(path)).absolute()

        root = f"{self._url.netloc}/{self._url.path}"
        if root.endswith("/"):
            root = os.path.join(root, os.path.basename(path))

        filenames = [subpath for subpath in path.glob("**/*") if subpath.is_file()] if path.is_dir() else [path]

        total = sum(filename.stat().st_size for filename in filenames)
        logger.info("%s file(s) (%sB) will be uploaded to %s", len(filenames), total, self._url)

        with Progress[int](
            total,
            unit="iB",
            desc="uploading to s3",
            sizeof_formatter=sizeof_fmt,
        ) as progress:
            for source_path in filenames:
                if source_path != path:
                    relpath = os.path.relpath(source_path, path)
                    target_path = os.path.join(root, relpath)
                else:
                    target_path = root

                with source_path.open("rb") as src, self._hf.open(target_path, "wb") as dst:
                    while True:
                        chunk = src.read(1024)
                        if not chunk:
                            break
                        dst.write(chunk)
                        progress.update(len(chunk))

    def delete(self) -> None:
        self._hf.rm(self._url.raw, recursive=True)

    def get_version(self) -> str | None:
        return str(self._hf.checksum(self._url.raw))

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[TextIO]:
        ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[BinaryIO]:
        ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[IO[Any]]:
        ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        decompress: DecompressOption = "none",
    ) -> ContextManager[IO[Any]]:
        @contextmanager
        def _open(
            mode: str,
            buffering: int,
            encoding: str | None,
            errors: str | None,
            newline: str | None,
        ) -> Iterator[IO[Any]]:
            if not self.exists():
                raise FileNotFoundError(self._url.raw)

            with self._hf.open(
                self._url.raw,
                mode,
                encoding=encoding,
                errors=errors,
                newline=newline,
                compression=None if decompress == "none" else "infer",
            ) as f:
                yield f

        return _open(mode, buffering, encoding, errors, newline)



from __future__ import annotations

import http.client
import os
import tempfile
import time
import urllib.parse
from contextlib import contextmanager
from os import PathLike
from typing import IO, Any, BinaryIO, Callable, ContextManager, Iterator, TextIO, overload

from minato.common import Progress
from minato.filesystems.filesystem import FileSystem
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode, sizeof_fmt, xopen


@FileSystem.register(["http", "https"])
class HttpFileSystem(FileSystem):
    def exists(self) -> bool:
        with self.http_head(self._url.raw, allow_redirects=True) as response:
            return response.status == 200

    def download(self, path: str | PathLike) -> None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        with open(path, "w+b") as fp:
            HttpFileSystem.http_get(self._url.raw, fp, allow_redirects=True)

    def delete(self) -> None:
        raise OSError("HttpFileSystem cannot delete files or directories.")

    def get_version(self) -> str | None:
        if not self.exists():
            raise FileNotFoundError(self._url.raw)

        with HttpFileSystem.http_head(self._url.raw, allow_redirects=True) as response:
            if response.status != 200:
                raise OSError(
                    "HEAD request failed for url {} with status code {}".format(self._url.raw, response.status)
                )
            return response.headers.get("ETag")

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[TextIO]:
        ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[BinaryIO]:
        ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[IO[Any]]:
        ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        decompress: DecompressOption = "none",
    ) -> ContextManager[IO[Any]]:
        @contextmanager
        def _open(
            mode: str,
            buffering: int,
            encoding: str | None,
            errors: str | None,
            newline: str | None,
        ) -> Iterator[IO[Any]]:
            if not self.exists():
                raise FileNotFoundError(self._url.raw)

            if "a" in mode or "w" in mode or "+" in mode or "x" in mode:
                raise ValueError("HttpFileSystem is not writable.")

            suffix = "-" + (self._url.path.split("/")[-1] if self._url.path else "")
            temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)

            try:
                HttpFileSystem.http_get(self._url.raw, temp_file, allow_redirects=True)
                temp_file.close()
                with xopen(
                    temp_file.name,
                    mode=mode,
                    encoding=encoding,
                    buffering=buffering,
                    errors=errors,
                    newline=newline,
                    decompress=decompress,
                ) as fp:
                    yield fp
            finally:
                os.remove(temp_file.name)

        return _open(mode, buffering, encoding, errors, newline)

    @staticmethod
    def http_get(
        url: str,
        temp_file: IO[Any],
        *,
        retries: int = 5,
        allow_redirects: bool = False,
    ) -> None:
        def get_response(connection: http.client.HTTPConnection, url: str) -> http.client.HTTPResponse:
            connection.request("GET", url)
            return connection.getresponse()

        url_parsed = urllib.parse.urlparse(url)
        with HttpFileSystem.http_request_with_retry(
            url_parsed,
            get_response,
            retries=retries,
            allow_redirects=allow_redirects,
        ) as response:
            content_length = response.headers.get("Content-Length")
            total = int(content_length) if content_length is not None else None

            with Progress[int](
                total,
                unit="iB",
                desc="downloading",
                sizeof_formatter=sizeof_fmt,
            ) as progress:
                while True:
                    chunk = response.read(1024)
                    if not chunk:
                        break
                    progress.update(len(chunk))
                    temp_file.write(chunk)

    @staticmethod
    def http_head(
        url: str,
        *,
        retries: int = 5,
        allow_redirects: bool = False,
    ) -> http.client.HTTPResponse:
        def get_response(connection: http.client.HTTPConnection, url: str) -> http.client.HTTPResponse:
            connection.request("HEAD", url)
            return connection.getresponse()

        url_parsed = urllib.parse.urlparse(url)
        response = HttpFileSystem.http_request_with_retry(
            url_parsed,
            get_response,
            retries=retries,
            allow_redirects=allow_redirects,
        )

        return response

    @staticmethod
    def http_request_with_retry(
        url_parsed: urllib.parse.ParseResult,
        get_response: Callable[[http.client.HTTPConnection, str], http.client.HTTPResponse],
        retries: int = 5,
        backoff_factor: int = 1,
        allow_redirects: bool = False,
    ) -> http.client.HTTPResponse:
        status_codes_to_retry = {502, 503, 504}
        status_codes_to_redirect = {300, 301, 302, 303, 307, 308}

        for retry in range(retries):
            try:
                connection = (
                    http.client.HTTPConnection(url_parsed.netloc)
                    if url_parsed.scheme == "http"
                    else http.client.HTTPSConnection(url_parsed.netloc)
                )
                path_with_query = url_parsed.path + ("?" + url_parsed.query if url_parsed.query else "")
                response = get_response(connection, path_with_query)

                if allow_redirects and response.status in status_codes_to_redirect:
                    redirect_url = response.headers.get("Location")
                    if redirect_url:
                        url_parsed = urllib.parse.urlparse(redirect_url)
                        continue
                elif response.status != 200:
                    if response.status in status_codes_to_retry and retry < retries - 1:
                        time.sleep(backoff_factor * (2**retry))
                        continue
                    else:
                        raise http.client.HTTPException(f"Request failed with status {response.status}")

                break
            except Exception as e:
                if retry < retries - 1:
                    time.sleep(backoff_factor * (2**retry))
                else:
                    raise e

        return response
